;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; f2c-si.el
;;; elisp code for loading stabs data from .si records.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Copyright (c) 1996 Harvey J. Stein <abel@netvision.net.il>, and
;; eventually <hjstein@netvision.net.il>
;; All Rights Reserved.
;; 
;; This package is covered by the GNU GPL.  You can freely use and
;; distribute it as long as it stays under the GNU GPL, and as long as
;; you distribute all the corresponding source code, and as long as this
;; message and the above copyright notice remains.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'parcil)
(require 'cl)
(require 'cl-19)

;;; full si record
(defun f2c-si-subname (f-decl)
  (nth 0 f-decl))

(defun f2c-si-filename (f-decl)
  (nth 1 f-decl))

(defun f2c-si-function-type (f-decl)
  (nth 2 f-decl))

(defun f2c-si-args (f-decl)
  (nth 3 f-decl))

(defun f2c-si-calls (f-decl)
  (nth 4 f-decl))

(defun f2c-si-locals (f-decl)
  (nth 5 f-decl))

(defun f2c-si-includes (f-decl)
  (nth 6 f-decl))

(defun f2c-si-externals (f-decl)
  (nth 7 f-decl))

(defun f2c-si-common (f-decl)
  (nth 8 f-decl))

(defun f2c-si-params (f-decl)
  (nth 9 f-decl))

;;; si argument block
(defun f2c-si-arg-name (arg-rec)
  (nth 0 arg-rec))

(defun f2c-si-conditioned-arg-type (arg-rec)
  (f2c-si-cnvt-double-stars (symbol-name (f2c-si-arg-type arg-rec))))

(defun f2c-si-cnvt-double-stars (arg)
  (if (string-match  "\\*\\*" arg)
      (concat (substring arg 0 (match-beginning 0))
	      "*(*)"
	      (substring arg (match-end 0)))
    arg))

(defun f2c-si-arg-type (arg-rec)
  (nth 1 arg-rec))

(defun f2c-si-arg-dimen (arg-rec)
  (if (not (null (cddr arg-rec)))
      (nth 2 arg-rec)
    nil))
  
;;; si include file block
(defun f2c-si-incs-file-name (inc-rec)
  (nth 0 inc-rec))

;;; si common block
(defun f2c-si-common-name (common)
  (nth 0 common))

(defun f2c-si-common-vars (common)
  (nth 1 common))

;;; si parameters
(defun f2c-si-param-name (param)
  (nth 0 param))

(defun f2c-si-param-value (param)
  (nth 1 param))

(defun f2c-si-inc->req (s)
  (if (symbolp s)
      s
    (intern s)))

(defun f2c-si-downcase-sym (s)
  (intern (downcase (symbol-name s))))

(defun f2c-si-si2stab (decl)
  (let* ((context (f2c-si-subname decl))
	 (common-vars (apply (function append)
			     (mapcar (function f2c-si-common-vars)
				  (f2c-si-common decl)))))
    (if (eq context '*undefined*)
	(setq context (f2c-si-filename decl)))
    (setq context (f2c-si-downcase-sym (f2c-si-inc->req context)))

    (insert ";;; Included files:\n")
    (mapcar (lambda (inc-file-rec)
	      (insert (format "(f2c-require '%s)\n"
			      (f2c-si-inc->req (f2c-si-incs-file-name inc-file-rec)))))
	 (f2c-si-includes decl))

    (insert "\n\n;;; Arguments:\n")
    (mapcar (lambda (var-rec)
	      (f2c-si-output-load-set context 'f2c-add-arg-var var-rec))
	    (f2c-si-args decl))
    
    (insert "\n\n;;; Local arrays:\n")
    (mapcar (lambda (var-rec)
		(if (not (member (f2c-si-arg-name var-rec)
				 common-vars))
		    (f2c-si-output-load-set context 
					 'f2c-add-local-var
					 var-rec)))
	    (f2c-si-arrays-only (f2c-si-locals decl)))
    
    (insert "\n\n;;; Common blocks:\n")
    (mapcar (lambda (common)
	      (let ((cname (f2c-si-common-name common)))
		(mapcar (lambda (cvar)
			  (f2c-si-output-load-set cname 
					'f2c-add-common-var
					(assoc cvar (f2c-si-locals decl))))
			(f2c-si-common-vars common))
		(insert (format "(f2c-add-subcontext '%s '%s)\n"
				context (f2c-si-downcase-sym cname)))))
	    (f2c-si-common decl))

    (insert "\n\n;;; Parameters:\n")
    (mapcar (lambda (param)
	      ;; Since parcil can't deal with character parameters, we
	      ;; skip them here...
	      (let ((type (assoc (f2c-si-param-name param)
				 (f2c-si-locals decl))))
		(if (or (not type)
			(not (string-match "^CHARACTER" (symbol-name (f2c-si-arg-type type)))))
		    (insert (format "(f2c-add-param '%s '%s %S)\n"
				    context
				    (f2c-si-downcase-sym (f2c-si-param-name param))
				    (f2c-si-param-value param))))))
	      (f2c-si-params decl))

    (insert "\n\n;;; Subcontexts (from include files):\n")
    (mapcar (lambda (inc-file-rec)
	      (insert (format "(f2c-add-subcontext '%s '%s)\n"
			      context
			      (f2c-si-inc->req (f2c-si-incs-file-name inc-file-rec)))))
	    (f2c-si-includes decl))))

(defun f2c-si-arrays-only (l)
  (cond ((null l) ())
	((f2c-si-arg-dimen (car l))
	 (cons (car l)
	       (f2c-si-arrays-only (cdr l))))
	(t (f2c-si-arrays-only (cdr l)))))


(defun f2c-si-output-load-set (fnam func var-decl)
  (let* ((var (downcase (symbol-name (f2c-si-arg-name var-decl))))
	 (dlist (f2c-si-arg-dimen var-decl))
	 (dimens (if dlist (mapcar (function downcase) dlist)
		     nil))
	 (dims (if dimens (f2c-si-si2stab-aref-convert dimens)
		   nil))
	 (fnam (downcase (symbol-name fnam))))
    (if (and dims (not (null dims)))
	(insert (format "(%s '%s '%s '%S '%S)\n"
			func fnam var (nth 0 dims) (nth 1 dims)))
      (insert (format "(%s '%s '%s)\n"
		      func fnam var)))))
	

(defun f2c-si-index-cleaner (strng)
  (if (string-match "^\\(.*\\):\\(.*\\)$" strng)
	(list (substring strng (match-beginning 2) (match-end 2))
	      (substring strng (match-beginning 1) (match-end 1)))
    (list strng "1")))

(defun f2c-si-si2stab-aref-convert (dimens)
  (f2c-si-transpose (mapcar (function f2c-si-index-cleaner) dimens)))

(defun f2c-si-transpose (l)
  (cond ((null l) ())
	((null (car l)) ())
	(t (cons (mapcar (function car) l)
		 (f2c-si-transpose (mapcar (function cdr) l))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun f2c-si-bufname (fname)
  (format "*f2c-si-load-%s*" fname))

(defun f2c-si-el-filename (fname)
  (let ((fname (file-name-nondirectory fname)))
    (if (string-match "\\.si$" fname)
	(concat (substring fname 0 (- (length fname) 3))
		".el")
      (concat fname ".el"))))

(defun f2c-si-load (fname &optional eval-data write-data)
  (when (file-exists-p fname)
    (let ((in-buffer (generate-new-buffer (f2c-si-bufname fname)))
	  (out-buffer (generate-new-buffer (f2c-si-el-filename fname)))
	  (oldbuf (current-buffer)))
      (unwind-protect
	  (save-excursion
	    (message "Loading SI file %s..." fname)
	    (set-buffer in-buffer)
	    (insert-file-contents fname)
	    (set-buffer out-buffer)
	    (let ((decl ()))
	      (while (condition-case ()
			 (progn (setq decl (read in-buffer))
				t)
		       (end-of-file ()))
		(f2c-si-si2stab decl)))
	    (if eval-data (eval-current-buffer))
	    (if write-data (write-file (f2c-si-el-filename fname)))
	    (message "Loading SI file %s...  Done." fname)))
      (kill-buffer in-buffer)
      (kill-buffer out-buffer))))

(provide 'f2c-si)
